from libifstate.util import logger, ipr
from libifstate.exception import LinkTypeUnknown
from abc import ABC, abstractmethod

class Link(ABC):
    _nla_prefix = 'IFLA_'
    _classes = {}

    def __new__(cls, *args, **kwargs):
        cname = cls.__name__
        if cname == Link.__name__:
            cname = "{}Link".format(kwargs["kind"].lower().capitalize())

        for c in Link.__subclasses__():
            if c.__name__ == cname:
                return super().__new__(c)

        raise LinkTypeUnknown()

    def __init__(self, name, **kwargs):
        self.cap_create = True
        self.settings = {
            'ifname': name,
        }
        self.settings.update(kwargs)
        self.attr_map = {
            'kind': ['IFLA_LINKINFO', 'IFLA_INFO_KIND'],
        }
        self.attr_idx = []
        self.idx = None

        if 'address' in self.settings:
            self.idx = next(iter(ipr.link_lookup(
                address=self.settings['address'])), None)
        if self.idx is None:
            self.idx = next(iter(ipr.link_lookup(ifname=name)), None)

        logger.debug('%s init %s', self.settings['ifname'], self.settings)

    def _drill_attr(self, data, keys):
        key = keys[0]
        d = data.get_attr(key)

        if d is not None:
            if len(keys) > 1:
                return self._drill_attr(d, keys[1:])
            else:
                return d
        else:
            return None

    def get_if_attr(self, key):
        if key in self.attr_map:
            return self._drill_attr(self.iface, self.attr_map[key])

        return self.iface.get_attr(self.name2nla(key))

    def commit(self):
        logger.debug('%s starting commit', self.settings['ifname'])

        # lookup for attributes requiring a interface index
        for attr in self.attr_idx:
            self.settings[attr] = next(iter(ipr.link_lookup(ifname=self.settings[attr])), self.settings[attr])

        if self.idx is not None:
            self.iface = next(iter(ipr.get_links(self.idx)), None)

            if self.cap_create and self.get_if_attr('kind') != self.settings['kind']:
                self.recreate()
            else:
                self.update()
        else:
            self.create()
        logger.debug('%s finished commit', self.settings['ifname'])

    def create(self):
        logger.debug('%s creating link', self.settings['ifname'])

        ipr.link('add', **(self.settings))

    def recreate(self):
        logger.debug('%s recreating link', self.settings['ifname'])
        ipr.link('del', index=self.idx)
        self.idx = None
        self.create()

    def update(self):
        logger.debug('%s updating link', self.settings['ifname'])

        old_state = self.iface['state']
        has_changes = False
        for setting in self.settings.keys():
            logger.debug('%s  %s: %s => %s', self.settings['ifname'], setting, self.get_if_attr(
                setting), self.settings[setting])
            if setting != 'kind' or self.cap_create:
                has_changes |= self.get_if_attr(
                    setting) != self.settings[setting]

        if has_changes:
            logger.debug('%s pending changes', self.settings['ifname'])
            if old_state:
                logger.debug('%s set state to down', self.settings['ifname'])
                ipr.link('set', index=self.idx, state='down')
                if not 'state' in self.settings:
                    self.settings['state'] = 'up'

            logger.debug('%s updating settings', self.settings['ifname'])
            ipr.link('set', index=self.idx, **(self.settings))
        else:
            logger.debug('%s no pending changes', self.settings['ifname'])

    def depends(self):
        return None

    @classmethod
    def name2nla(self, name):
        '''
        Convert human-friendly name into NLA name

        Example: address -> IFLA_ADDRESS

        Requires self.prefix to be set
        '''
        name = name.upper()
        if name.find(self._nla_prefix) == -1:
            name = "%s%s" % (self._nla_prefix, name)
        return name



from libifstate.util import logger
from abc import ABC, abstractmethod

class IfAttrList(list):
    def __init__(self, l):
        for i, v in enumerate(l):
            if isinstance(v, dict):
                l[i] = IfAttrDict(v)
            elif isinstance(v, list):
                l[i] = IfAttrList(v)
        super().__init__(l)

class IfAttrDict(dict):
    def __init__(self, d):
        for k, v in d.items():
            if isinstance(v, dict):
                d[k] = IfAttrDict(v)
            elif isinstance(v, list):
                d[k] = IfAttrList(v)
        super().__init__(d)

    def __getitem__(self, key):
        # if not key in self:
        #     print("booom")
        return super().__getitem__(key)

class Parser(ABC):
    _default_ifstates = {
        'ignore': [
            r'^docker\d+',
            r'^lo$',
            r'^ppp\d+$',
            r'^veth',
            r'^virbr\d+',
        ],
        'interfaces': {}
    }

    @abstractmethod
    def __init__(self, name, **kwargs):
        self.ifstate = {}
        pass

    def config(self):
        return IfAttrDict({**self._default_ifstates, **self.ifstates})



from libifstate.exception import LinkDuplicate
from libifstate.link import Link
from libifstate.parser import Parser
from libifstate.util import logger, ipr
import re

class IfState():
    def __init__(self):
        self.links = {}
        self.ignore = set([])
    
    def update(self, ifstates):
        # add interfaces from config
        for ifstate in ifstates['interfaces']:
            name = ifstate.pop('name')
            if name in self.links:
                raise LinkDuplicate()
            self.links[name] = Link(name, **ifstate)

        # add ignore list items
        self.ignore.update(ifstates['ignore'])

    def commit(self):
        last = 0
        commited = []
        while len(commited) < len(self.links):
            for name, link in self.links.items():
                dep = link.depends()
                if dep is None or dep in commited:
                    link.commit()
                    commited.append(name)
            if last == len(commited):
                raise LinkCircularLinked()

        for link in ipr.get_links():
            name = link.get_attr('IFLA_IFNAME')
            # skip links on ignore list
            if not name in self.links and not any(re.match(regex, name) for regex in self.ignore):
                info = link.get_attr('IFLA_LINKINFO')
                # remove virtual interface
                if info is not None:
                    kind = info.get_attr('IFLA_INFO_KIND')
                    logger.info('%s is a orphan %s interface => remove', name, kind or 'virtual')
                    ipr.link('del', index=link.get('index'))
                # shutdown physical interfaces
                else:
                    if link.get('state') == 'down':
                        logger.warning('%s is a orphan physical interface', name)
                    else:
                        logger.warning('%s is a orphan physical interface => shutdown', name)
                        ipr.link('set', index=link.get('index'), state='down')

    def template(self):
        ifs_links = []
        for ipr_link in ipr.get_links():
            name = ipr_link.get_attr('IFLA_IFNAME')
            # skip links on ignore list
            if not any(re.match(regex, name) for regex in Parser._default_ifstates['ignore']):
                kind = 'physical'

                info = ipr_link.get_attr('IFLA_LINKINFO')
                if info is not None:
                    kind = info.get_attr('IFLA_INFO_KIND')

                ifs_link = {
                        'ifname': name,
                        'kind': kind,
                }

                # cname = "{}Link".format(kind.lower().capitalize())
                # for c in Link.__subclasses__():
                #     if c.__name__ == cname:

                ifs_links.append(ifs_link)

        return {
            'interfaces': ifs_links,
        }



from setuptools import setup, find_packages
from datetime import date


def readme():
    with open("README.md") as f:
        return f.read()


def requirements():
    req = []
    with open("requirements.txt") as fd:
        for line in fd:
            line.strip()
            if not line.startswith("#"):
                req.append(line)
    return req


setup(
    name="ifstate",
    version='0.1',
    description="manage host interface settings in a declarative manner",
    author="Thomas Liske",
    author_email="thomas@fiasko-nw.net",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/liske/ifstate",
    license="GPL3+",
    packages=find_packages(),
    install_requires=requirements(),
    entry_points={
        "console_scripts": ["ifstate = ifstate.ifstate:main"]
    },
)



from libifstate.link.base import Link
from libifstate.exception import LinkCannotAdd

class PhysicalLink(Link):
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        self.cap_create = False

    def create(self):
        raise LinkCannotAdd()



from libifstate.link.base import Link

class VlanLink(Link):
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        self.attr_map.update({
            'vlan_id': ['IFLA_LINKINFO', 'IFLA_INFO_DATA', 'IFLA_VLAN_ID'],
            'vlan_protocol': ['IFLA_LINKINFO', 'IFLA_INFO_DATA', 'IFLA_VLAN_PROTOCOL'],
        })
        self.attr_idx.append('link')

    def depends(self):
        return self.settings['link']



#!/usr/bin/env python3

from libifstate.parser import YamlParser
from libifstate import IfState
from libifstate.util import logger
from collections import namedtuple

import argparse
import logging
import yaml

class Actions():
    CHECK = "check"
    CONFIG = "config"
    TEMPLATE = "template"

def main():
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-d", "--debug", action="store_true")
    group.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("action", choices=list(a.lower() for a in dir(Actions) if not a.startswith('_')))

    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.INFO)
    elif args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)

    ifs = IfState()

    if args.action == Actions.TEMPLATE:
        print(yaml.dump(ifs.template()))

    if args.action in [Actions.CHECK, Actions.CONFIG]:
        parser = YamlParser('./test.yml')
        ifstates = parser.config()

        ifs.update(ifstates)

        if args.action == Actions.CHECK:
            pass
        else:
            ifs.commit()

if __name__ == "__main__":
    main()



from libifstate.util import logger
from libifstate.parser.base import Parser
import yaml

class YamlParser(Parser):
    def __init__(self, fn):
        logger.debug('YamlParser parsing %s', fn)
        with open(fn) as fh:
            self.ifstates = yaml.full_load(fh)
